import asyncio
from abc import abstractmethod
from inspect import iscoroutinefunction as is_coroutine
from .states import State, StepStates


class BaseStep(StepStates):
    async def do_run(self, *args, **kwargs):
        self._state = State.RUNNING
        try:
            res = await self.run(*args, **kwargs)
        except Exception as error:
            self._state = State.FAILURE
            raise error
        self._state = State.SUCCESS
        return res

    async def do_compensate(self, *args, **kwargs):
        self._state = State.RUNNING
        res = await self.compensate(*args, **kwargs)
        self._state = State.COMPENSATED
        return res

    @abstractmethod
    async def run(self, *args, **kwargs):
        pass

    @abstractmethod
    async def compensate(self, *args, **kwargs):
        pass


class Step(BaseStep):
    """
    Creates an saga step with a function to call with its compensation
    """

    def __init__(self, func, compensation=None):
        assert callable(func) is True, "'func' argument must be callable"
        self.func = func
        if compensation:
            assert callable(compensation) is True, "'func' argument must be callable"
            self.compensation = compensation

        super().__init__()

    async def _call_func(self, func, *args, executor=None, **kwargs):
        if is_coroutine(func):
            return await func(*args, **kwargs)
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(executor, func, *args, **kwargs)

    async def run(self, *args, **kwargs):
        return await self._call_func(self.func, *args, **kwargs)

    async def compensate(self, *args, **kwargs):
        return (
            await self._call_func(self.compensation, *args, **kwargs)
            if self.compensation
            else None
        )



import asyncio
import pytest


@pytest.fixture(scope='module')
def event_loop():
    """Fixture for the async event loop

    Yields:
        event loop
    """
    loop = asyncio.get_event_loop()
    yield loop
    loop.close()



from collections.abc import Sequence
from .event_gather import do_gather
from .states import State, StepStates
from .step import Step
from typing import List


class Saga(StepStates):
    steps: List[Step] = []

    def __init__(self, steps: list = None):
        if steps is None:
            steps = []
        assert isinstance(steps, Sequence), "Steps must be a sequence"
        self.steps = steps
        super().__init__()

    async def run(self, *args, executor=None, exceptions=None, **kwargs):
        futures = [step.do_run(*args, **kwargs) for step in self.steps]
        await do_gather(*futures, allowed_exc=(exceptions))
        failure = any([step.failure for step in self.steps])
        if failure:
            futures = [
                step.do_compensate(*args, **kwargs)
                for step in self.steps
                if step.processed
            ]
            await do_gather(*futures, allowed_exc=(exceptions))
        self._state = State.FAILURE if failure else State.SUCCESS


class SagaBuilder(object):
    def __init__(self):
        self.steps = []

    @staticmethod
    def create():
        return SagaBuilder()

    def add_step(self, func, compensation):
        step = Step(func, compensation)
        self.steps.append(step)
        return self

    def build(self):
        return Saga(self.steps)



from sagas import Saga, SagaBuilder, Step


class BaseTestCase(object):
    def assertEqual(self, first, second, msg=None):
        """Fail if the two objects are unequal as determined by the '=='
           operator.
        """
        assert first == second, msg


class TestSaga(BaseTestCase):
    def test_run_step_succeeds(self, event_loop):
        counters = [0, 10]

        async def action():
            counters[0] += 1

        step_1 = Step(action)
        event_loop.run_until_complete(
            Saga([step_1]).run(exceptions=(OSError, Exception))
        )
        self.assertEqual(counters[0], 1)

    def test_multiple_steps_succeed(self, event_loop):
        counters = [0, 0, 0, 0]

        def action(index):
            counters[index] += 1

        step_0 = Step(lambda: action(0))
        step_1 = Step(lambda: action(1))
        step_2 = Step(lambda: action(2))
        step_3 = Step(lambda: action(3))

        event_loop.run_until_complete(
            Saga([step_0, step_1, step_2, step_3]).run(exceptions=(OSError))
        )
        self.assertEqual(counters, [1, 1, 1, 1])

    def test_step_fails(self, event_loop):
        counters = [0]

        async def action():
            counters[0] += 1
            raise Exception("test_step_fails")

        step_1 = Step(action)
        try:
            event_loop.run_until_complete(Saga([step_1]).run(exceptions=(OSError)))
        except Exception as error:
            assert isinstance(error, Exception) is True

    def test_compensation_succeeds(self, event_loop):
        counters = [0]

        async def action():
            counters[0] += 1
            raise Exception("test_step_fails")

        async def comp():
            counters[0] -= 1

        step_1 = Step(action, comp)
        event_loop.run_until_complete(
            Saga([step_1]).run(exceptions=(OSError, Exception))
        )
        self.assertEqual(counters[0], 0)


class TestSagaBuilder(BaseTestCase):
    def test_run_and_compensate(self, event_loop):
        counters = [0, 10]

        async def action():
            counters[0] += 1

        async def comp():
            counters[0] -= 1

        saga = SagaBuilder.create().add_step(action, comp).build()
        event_loop.run_until_complete(saga.run(exceptions=(OSError)))
        self.assertEqual(counters[0], 1)



import asyncio


async def do_gather(*coroutines, allowed_exc):
    pending = list(map(asyncio.ensure_future, coroutines))
    try:
        await asyncio.gather(*pending)
    except allowed_exc:
        for t in pending:
            t.cancel()



import asyncio
from sagas import SagaBuilder

counter = [10, 20]
amount = 30


def add(amount, index):
    """Increments array index by amount

    Args:
        amount (int): increment value
        index (int): index of value to increment
    """
    counter[index] += amount
    raise Exception("error occurred")


def decrement(amount, index):
    """Decrements array index by amount

    Args:
        amount (int): decrement value
        index (int): index of value to decrement
    """
    counter[index] -= amount
    return counter


saga_builder = SagaBuilder.create()

saga = (
    saga_builder.add_step(lambda: add(amount, 0), lambda: decrement(amount, 0))
    .add_step(lambda: add(amount, 1), lambda: decrement(amount, 1))
    .build()
)

if __name__ == "__main__":
    asyncio.run(saga.run(exceptions=(OSError, Exception)))
    print(counter)  # [10, 20]



from enum import Enum


class State(Enum):
    SUCCESS = 0
    COMPENSATED = 1
    FAILURE = 2
    CREATED = 3
    RUNNING = 4


class StepStates:
    def __init__(self):
        self._state = State.CREATED

    @property
    def state(self):
        return self._state

    @property
    def complete(self):
        return self._state in [State.COMPENSATED, State.SUCCESS, State.FAILURE]

    @property
    def success(self):
        return self._state in [State.COMPENSATED, State.SUCCESS]

    @property
    def processed(self):
        return self._state in [State.FAILURE, State.SUCCESS]

    @property
    def failure(self):
        return self._state == State.FAILURE
